/*
 * Metronome_pz2
 *
 * Copyright (C) 2005 Jannis Achstetter (jannis),
 * original metronome (for pz0):
 * Copyright (C) 2005 Stephen Lawrence (Iklop)
 *
 * Please don't blame me for lame, slow or ugly code,
 * I'm just starting to learn C. Send patches if you want
 * and I will learn and do better next time.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <string.h>
#include "pz.h"

/* --- VARS ---- */
static PzModule *module;		//opaque type, used only to fetch path info
//sound
static pz_dsp_st dspz;
//metronome
static ttk_timer metronome_timer_id;	//event timer
static struct timeval metronome_goal;	//goal time
static int metronome_bpm = 120;		//beats per minute
static int metronome_mspb = 500;	//milliseconds per beat
static int metronome_volume;		//volume
static int metronome_scale = 1;		//scale
static int metronome_scale2b = 1;	//scale to be, this keeps the main main always on track
static int metronome_tick_count = 0;	//keeps track of ticks
char extra[16];				//word underneath the speed
//tap varialbes
#define METRONOME_TAP_SIZE	4
static int metronome_tap_diffs[METRONOME_TAP_SIZE];
static int metronome_tap_count;
static struct timeval metronome_tap_time;//tap time
//modes & submodes
#define METRONOME_MODE_NORMAL	0
#define METRONOME_MODE_TAP		1
#define METRONOME_MODE_LAST		1
#define METRONOME_SUBMODE_NORMAL	0
#define METRONOME_SUBMODE_VOLUME	1
#define METRONOME_SUBMODE_SCALE		2
#define METRONOME_SUBMODE_LAST		2
//mode variables
static int metronome_mode = METRONOME_MODE_NORMAL;
static int metronome_submode = METRONOME_SUBMODE_NORMAL;
//state variables
static int metronome_paused = 0;	//paused state
static int metronome_piezo = 0;		//piezo state
//speed limits
#define METRONOME_MAX	208 		//maximum bpm
#define METRONOME_MIN 	40		//minimum bpm

PzWidget *wid_glob;
ttk_surface srf_glob;
int opt_tick = 0;			//optical tick state
#define OPT_DISP_TIME 100		//how long black squares will be displayed (ms)

/* --- Prototypes --- */
void init_metronome ();
PzWindow *new_metronome_window();
void draw_metronome (PzWidget *wid, ttk_surface srf);
int event_metronome (PzEvent *ev);
void metronome_mode_changed();
void crt_extra ();
void draw_mode_info(PzWidget *wid, ttk_surface srf, int y);
void metronome_create_tick();
void metronome_delete_tick();
void metronome_tick(int level);
void clear_opt_tick();
void metronome_check_ticks();
int metronome_get_tap_mspb();
void metronome_tap();
void cleanup_metronome();
void unload_metronome();

/* --- Implementation --- */
/********************************************************************/
/********************************************************************/

//INIT (is called at after pz2-startup)
void init_metronome()
{
	module = pz_register_module ("metronome", unload_metronome);
	pz_menu_add_action ("/Extras/Utilities/Metronome", new_metronome_window);
}

//WINDOW-Creator (when selected on menu)
PzWindow *new_metronome_window()
{
	//set the goal to zero
	metronome_goal.tv_sec = 0;
	metronome_goal.tv_usec = 0;
	//set the timer which checks ticks
	metronome_timer_id = ttk_create_timer(1,metronome_check_ticks);
	//setup sound
	pz_dsp_setup(&dspz, 1, 44100);
	pz_dsp_open(&dspz, PZ_DSP_LINEOUT);
	pz_dsp_setup(&dspz, 1, 44100);
	metronome_volume = pz_dsp_get_volume( &dspz );
	//create us a window
	PzWindow *ret;
	ret = pz_new_window ("Metronome", PZ_WINDOW_NORMAL);
	pz_add_widget (ret, draw_metronome, event_metronome);
	return pz_finish_window (ret);
}

void draw_metronome (PzWidget *wid, ttk_surface srf)
{
	if (opt_tick!=0) {
		if (opt_tick==1) {
			ttk_fillrect(srf,wid->w-30,0,wid->w,30,ttk_makecol (BLACK));
		} else {
			ttk_fillrect(srf,0,0,20,20,ttk_makecol (BLACK));
		}
	}
	char speedtext[20];
	sprintf(speedtext,"%d BPM; %d msec",metronome_bpm,metronome_mspb);
	int y = wid->y + 10 ;
	ttk_text (srf, ttk_textfont, wid->x + (wid->w - ttk_text_width (ttk_textfont, speedtext)) / 2, y, ttk_makecol (BLACK), speedtext);
	y += 12;
	crt_extra();
	ttk_text (srf, ttk_textfont, wid->x + (wid->w - ttk_text_width (ttk_textfont, extra)) / 2, y, ttk_makecol (BLACK), extra);
	y += 30;
	draw_mode_info(wid, srf, y);
	char *bottomtext = (char *)strdup ("Press 'menu' to exit.");
	ttk_text (srf, ttk_textfont, wid->x + (wid->w - ttk_text_width (ttk_textfont, bottomtext)) / 2, wid->y + wid->h - ttk_text_height (ttk_textfont) - 5, ttk_makecol (BLACK), bottomtext);
	wid_glob = wid;
	srf_glob = srf;
}

int event_metronome (PzEvent *ev)
{
	switch (ev->type) {
	case PZ_EVENT_BUTTON_DOWN: //button DOWN (better for tap)
		if (ev->arg == PZ_BUTTON_ACTION && metronome_mode == METRONOME_MODE_TAP) {
			metronome_tick(0);
			metronome_tap();
		}
		break;
	case PZ_EVENT_BUTTON_UP:  	//button released
		switch (ev->arg) {
		case PZ_BUTTON_MENU:	//will exit
			cleanup_metronome();
			pz_close_window (ev->wid->win);
			break;
		case PZ_BUTTON_PLAY:	//toggle pause
			metronome_paused = !metronome_paused;
			ev->wid->dirty++;
			break;
		case PZ_BUTTON_PREVIOUS://toggle piezo
			metronome_piezo = !metronome_piezo;
			break;
		case PZ_BUTTON_NEXT:	//toggle MAIN (tap<>tick) mode
			if (metronome_mode != METRONOME_MODE_LAST) {
				metronome_mode++;
			} else {
				metronome_mode = METRONOME_MODE_NORMAL;
			}
			metronome_mode_changed();
			ev->wid->dirty++;
			break;
		case PZ_BUTTON_ACTION:	//switch submodes or tap in
			if (metronome_mode == METRONOME_MODE_NORMAL) {
				if (metronome_submode != METRONOME_SUBMODE_LAST) {
					metronome_submode++;
				} else {
					metronome_submode = METRONOME_SUBMODE_NORMAL;
				}
			}
			ev->wid->dirty++;
			break;
		}
	case PZ_EVENT_DESTROY:
		break;
	case PZ_EVENT_SCROLL:
		switch (metronome_submode) {
		case METRONOME_SUBMODE_NORMAL:
			if (ev->arg > 0 && metronome_bpm < METRONOME_MAX) {
				metronome_bpm++;
				metronome_mspb = 60000/metronome_bpm;
			} else if (ev->arg < 0 && metronome_bpm> METRONOME_MIN) {
				metronome_bpm--;
				metronome_mspb = 60000/metronome_bpm;
			}
			break;
		case METRONOME_SUBMODE_VOLUME:
			if (ev->arg > 0 && metronome_volume < 98) {
				pz_dsp_vol_change(&dspz,3);
				metronome_volume += 3;
			} else if (ev->arg < 0 && metronome_volume > 2) {
				pz_dsp_vol_change(&dspz,-3);
				metronome_volume -= 3;
			}
			break;
		case METRONOME_SUBMODE_SCALE:
			if (ev->arg > 0 && metronome_scale2b < 4) {
				metronome_scale2b++;
			} else if (ev->arg < 0 && metronome_scale2b > 1) {
				metronome_scale2b--;
			}
			break;
		}
		ev->wid->dirty++;
	}
	/* Normal event return - handled, no clicky noise */
	return 0;
}

//Clears some VARs on mode-changes
void metronome_mode_changed()
{
	int i = 0;
	switch (metronome_mode) {
	case METRONOME_MODE_NORMAL:
		//reset the goal timer
		metronome_goal.tv_sec = 0;
		metronome_goal.tv_usec = 0;
		metronome_tick_count = 0;
		metronome_mspb = 60000/metronome_bpm;
		break;
	case METRONOME_MODE_TAP:
		//clear the tap array
		for( i=0; i<METRONOME_TAP_SIZE; i++ )
		metronome_tap_diffs[i] = 0;
		//reset the tap timer
		metronome_tap_time.tv_sec = 0;
		metronome_tap_time.tv_usec = 0;
		metronome_tap_count = 0; //set the tap count to zero
		break;
	}
}

//Creates string to print underneath speed-value
void crt_extra ()
{
	if( metronome_paused ) {
		sprintf( extra, "%s", "paused" );
	} else if( metronome_bpm<=60 ) {
		sprintf( extra, "%s", "largo" );
	} else if( metronome_bpm <= 66 ) {
		sprintf( extra, "%s", "larghetto" );
	} else if( metronome_bpm <= 76 ) {
		sprintf( extra, "%s", "adagio" );
	} else if( metronome_bpm <= 108 ) {
		sprintf( extra, "%s", "andante" );
	} else if( metronome_bpm <= 120 ) {
		sprintf( extra, "%s", "moderato" );
	} else if( metronome_bpm <= 168 ) {
		sprintf( extra, "%s", "allegro" );
	} else if( metronome_bpm <= 200 ) {
		sprintf( extra, "%s", "presto" );
	} else {
		sprintf( extra, "%s", "prestissimo" );
	}
}

//draws the specific mode-informations (vol-bar, used mode) at widget wid, surface srf, pos y
void draw_mode_info(PzWidget *wid, ttk_surface srf, int y)
{
	if( metronome_mode == METRONOME_MODE_TAP) {
		char *modetext = (char *)strdup ("TAP mode");
		ttk_text (srf, ttk_textfont, wid->x + (wid->w - ttk_text_width (ttk_textfont, modetext)) / 2, y, ttk_makecol (BLACK), modetext);
	} else if (metronome_submode == METRONOME_SUBMODE_NORMAL) {
		char *modetext = (char *)strdup ("SPEED mode");
		ttk_text (srf, ttk_textfont, wid->x + (wid->w - ttk_text_width (ttk_textfont, modetext)) / 2, y, ttk_makecol (BLACK), modetext);
	} else if (metronome_submode == METRONOME_SUBMODE_VOLUME) {
		int volume_length = ((wid->w - 24) * metronome_volume) / 110;
		ttk_rect (srf, 10, y, wid->w - 22 , y + 11, ttk_makecol (BLACK));
		ttk_fillrect (srf, 12, y+2, 11+volume_length, y+9, ttk_makecol (BLACK));
	} else if (metronome_submode == METRONOME_SUBMODE_SCALE) {
		char *modetext = (char *)strdup ("");
		sprintf(modetext,"SCALE: %d",metronome_scale2b);
		ttk_text (srf, ttk_textfont, wid->x + (wid->w - ttk_text_width (ttk_textfont, modetext)) / 2, y, ttk_makecol (BLACK), modetext);
	}
}

//TICK Sound
typedef struct wftable {
	int nsamples;		/* number of samples in the next array */
	int samples[282];	/* array of 0..128 waveform samples */
				/* these are to be played at 44.1khz */
} wftable;

static wftable metronome_tick_waveform = {  35, { 127,127,128,129,133,139,148,159,169,170,161,146,122,99,80,73,78,89,100,109,117,118,119,118,115,114,115,122,132,146,157,160,157,147,136 }};
static short * metronome_tick_buffer;

//Creates the buffer used to make the tick sound
void metronome_create_tick()
{
	int bp;
	int f1p = 0;
	//nsamps = ontime * 44100 / 1000;
	metronome_tick_buffer = (short *)malloc( sizeof( short ) * 44 );
	//play the two tones
	for( bp = 0 ; bp < 44; bp++ )
	{
		//copy over and add the two waveforms
		metronome_tick_buffer[bp] = ( metronome_tick_waveform.samples[f1p++] );
		metronome_tick_buffer[bp] = metronome_tick_buffer[bp] << 6;
		//adjust sample pointers
		if( f1p >= metronome_tick_waveform.nsamples )
			f1p = 0;
	}
}

//Deletes the tick sound
void metronome_delete_tick()
{
	free( metronome_tick_buffer );
}

//Makes a sound
void metronome_tick( int level )
{
	int goal_vol, cur_vol, delta;
	cur_vol = pz_dsp_get_volume(&dspz);
	if( level==0 ) goal_vol = metronome_volume;
	else goal_vol = metronome_volume - 15;
	delta = goal_vol - cur_vol;
	pz_dsp_vol_change( &dspz, delta );
	if( metronome_piezo )
	{
		if( level==0 ) ipod_beep();
	}
	else
	{
#ifdef IPOD
		pz_dsp_write( &dspz, metronome_tick_buffer, 44 * sizeof(short));
#endif
	}
	if (level==0) opt_tick = 1; else opt_tick = 2;
	wid_glob->dirty++;
	ttk_create_timer(OPT_DISP_TIME,clear_opt_tick);
}

void clear_opt_tick()
{
	opt_tick = 0;
	wid_glob->dirty++;
}

//Determines whether a tick sound should be made
void metronome_check_ticks()
{
	if (!metronome_paused && metronome_mode!=METRONOME_MODE_TAP) {
		struct timeval now;
		int secdiff;
		int usecdiff;
		int diff;
		gettimeofday( &now, NULL );
		//set the goal to now if the goal is zero
		if( metronome_goal.tv_sec==0 && metronome_goal.tv_usec==0 )
		{
			metronome_goal = now;
			diff = 0;
		}
		else
		{
			//calculate the difference between now and the goal
			secdiff = (metronome_goal.tv_sec - now.tv_sec) * 1000;
			usecdiff = (metronome_goal.tv_usec - now.tv_usec ) / 1000;
			diff = secdiff + usecdiff;
		}
		if( diff <= 0 )
		{
			int usec = 0;

			//make a sound
			metronome_tick( metronome_tick_count );
			if( metronome_tick_count==0 )
			{
				metronome_scale = metronome_scale2b;
			}
			//increase the tick count
			if( metronome_tick_count >= metronome_scale - 1 )
				metronome_tick_count = 0;
			else
				metronome_tick_count++;
			//calculate the next time there should be a tick
			usec = (metronome_goal.tv_usec + (metronome_mspb * 1000 / metronome_scale));
			metronome_goal.tv_sec += usec/1000000;
			metronome_goal.tv_usec = usec%1000000;
		}
	}
	metronome_timer_id = ttk_create_timer(1,metronome_check_ticks);
}

//Calculates the average milliseconds per beat
int metronome_get_tap_mspb()
{
	int loop;
	int diff;
	int sum = 0;
	for( loop=0; loop<METRONOME_TAP_SIZE; loop++ )
	{
		diff = metronome_tap_diffs[loop];
		if( diff == 0 )
			break;
		sum += diff;
	}
	if( loop==0 )
		return 0;
	return (sum / loop);
}

//Calculates the difference between taps in tap mode
void metronome_tap()
{
	struct timeval now;
	gettimeofday( &now, NULL );
	//make sure this isn't the first tap so a difference can be calculated
	if( metronome_tap_time.tv_sec!=0 || metronome_tap_time.tv_usec!=0 )
	{
		int sec;
		int usec;
		int diff;
		int mspb;
		int bpm;
		//get difference between previous tap and now
		sec = (now.tv_sec - metronome_tap_time.tv_sec) * 1000;
		usec = (now.tv_usec - metronome_tap_time.tv_usec) / 1000;
		diff = sec + usec;
		//add difss to diff array
		metronome_tap_diffs[metronome_tap_count] = diff;
		//set the tap count/position in array
		if( metronome_tap_count == METRONOME_TAP_SIZE-1 )
			metronome_tap_count = 0;
		else
			metronome_tap_count++;
		//get the average milliseconds per tap beat & bpm
		mspb = metronome_get_tap_mspb();
		bpm = 60000/mspb;
		//limit the bpm
		if( bpm > METRONOME_MAX )
			bpm = METRONOME_MAX;
		else if( bpm < METRONOME_MIN )
			bpm = METRONOME_MIN;
		//draw the bpm
                metronome_mspb = mspb;
		metronome_bpm = bpm;
	}
	//set the metronome tap time to now
	metronome_tap_time = now;
}

//Module cleanup (when exiting Metronome)
void cleanup_metronome()
{
	ttk_destroy_timer(metronome_timer_id);
	metronome_delete_tick();
	pz_dsp_close(&dspz);
}

//Module unloading (when exiting PZ)
void unload_metronome()
{
}

//INIT-function-caller
PZ_MOD_INIT(init_metronome)    //No semicolon at end of line
